/**************************************************************
 *
 * File: hw3_soln.c
 * Author: Steve Wolfman
 *
 * Description: Sample solution to CSE142 01sp project #3.
 * See http://www.cs.washington.edu/142/ (navigate to hw#3, 01sp) for
 * description of assignment. 
 *
 **************************************************************/

#include <stdio.h>   /* For I/O.     */
#include <assert.h>  /* For assert.  */
#include <math.h>    /* For sqrt.    */
#include <stdlib.h>  /* For exit.    */
#include <limits.h>  /* For INT_MAX. */

/********* Useful constants **************/
#define BASE              36  /* The numerical base used for "words"
                                 in the program. */
#define WORD_NUM_SENTINEL -1  /* Value returned by getWordNum if a
                                 terminating word is read. (NOT AN
                                 ERROR CODE!) */

/********* Function prototypes ***********/
void printHeader(void);           /* Prints a decorative header.      */

void printFooter(void);           /* Prints a decorative footer.      */

int  getWordNum(void);            /* Reads a word, returns its value. */

char getDigit(void);              /* Reads and returns a digit.       */

int  isValidDigit(char digit);    /* Checks a digit against the BASE. */

int  addDigit(int currNum,        /* Given current word value, finds  */
              char digit);        /* the value with digit tacked on.  */

int  getNumFromDigit(char digit); /* Gets numerical value of digit.   */

int  isPrime(int num);            /* Checks if num is prime.          */

int  isDivisibleBy(int num,       /* Checks if divisor divides num    */
                   int divisor);  /* evenly.                          */

void printAnswer(int bIsPrime);   /* Outputs correct message based    */
                                  /* on answer (prime or not).        */

/********* Main function *****************/
int main(void)
{
  int wordNum;
  int bIsPrime;

  printHeader();
  
  /* Read words and output primality until getWordNum finds the
     sentinel. */
  wordNum = getWordNum();
  while (wordNum != WORD_NUM_SENTINEL) {
    /* Note: 1 is neither prime nor composite.
       I'll just treat it as composite since there is no spec for what
       to do with it otherwise. */
    bIsPrime = isPrime(wordNum);
    printAnswer(bIsPrime);

    /* loop update */
    wordNum = getWordNum();
  }

  printFooter();
  
  /* Report success. */
  return 0;
}

/*********************************************************************
 * printHeader:
 *   Prints a pretty header for the program.
 *
 *precondition: none
 *postcondition: printed header
 *********************************************************************/
void printHeader(void)
{
  printf("Welcome to the prime word tester.\n");
  printf("Testing words in base %d.\n", BASE);
  printf("Enter a series of words.\n");
  printf("One per line, please.\n");
}

/*********************************************************************
 * printFooter:
 *   Prints a pretty footer for the program.
 *
 *precondition: none
 *postcondition: printed footer
 *********************************************************************/
void printFooter(void)
{
  printf("Stopping.\n");
}


/*********************************************************************
 * printAnswer:
 *   Prints an appropriate message depending on the answer passed.
 *
 *     bIsPrime: holds whether the number was prime.
 *
 *precondition: bIsPrime is a Boolean (0 or 1)
 *postcondition: prints appropriate message for bIsPrime
 *********************************************************************/
void printAnswer(int bIsPrime)
{
  assert(bIsPrime == 0 || bIsPrime == 1);
  
  if (bIsPrime) {
    printf("PRIME\n");
  }
  else {
    printf("composite\n");
  }
}

/*********************************************************************
 * getWordNum
 *   Reads a word from standard input and returns its base 36
 *   representation. If the word starts with a 0, immediately returns
 *   WORD_NUM_SENTINEL. Exits on malformed or overly large words.
 *
 *precondition: none
 *postcondition: returns numerical value of the next word,
 *               WORD_NUM_SENTINEL, or exits program as appropriate.
 *********************************************************************/
int  getWordNum(void)
{
  char currDigit;          /* Current digit in the word. */
  int wordValue = 0;       /* Word's value so far.       */

  currDigit = getDigit();

  /* Check for a terminating word. */
  if (currDigit == '0') {
    return WORD_NUM_SENTINEL;
  }

  /* Build up the word's value. */
  while (currDigit != '\n') {
    wordValue = addDigit(wordValue, currDigit);

    /* loop update */
    currDigit = getDigit();
  }

  return wordValue;
}

/*********************************************************************
 * getDigit
 *   Reads a digit from standard input and returns it. Exits on
 *   malformed digits.
 *
 *precondition: none
 *postcondition: returns a digit 0-9 or A-Z (depending on BASE) or \n.
 *********************************************************************/
char getDigit(void)
{
  char digit;

  if (scanf("%c", &digit) != 1) {
    printf("Error reading digit.\n");
    exit(EXIT_FAILURE);
  }
  
  if (!isValidDigit(digit) && digit != '\n') {
    printf("Invalid character: '%c'.\n", digit);
    exit(EXIT_FAILURE);
  }
  
  return digit;
}

/*********************************************************************
 * isValidDigit:
 *   Checks whether the digit is valid in the current BASE.
 *
 *     digit: the digit to check
 *
 *precondition: none
 *postcondition: returns 1 if digit is valid, 0 otherwise
 *********************************************************************/
int isValidDigit(char digit)
{
  int tenDigit,      /* Is a normal base 10 digit?                    */
      alphaDigit;    /* Is a "letter" type digit (A-Z)?               */

  /* Check if it's a ten digit AND it's legal in the current base. */
  tenDigit = ((digit >= '0' && digit <= '9') &&
              ((int)digit - (int)'0' < BASE));

  /* Check if it's an alpha digit AND legal in the current base.   */
  alphaDigit = ((digit >= 'A' && digit <= 'Z') &&
                ((int)digit - (int)'A' + 10 < BASE));
  
  return tenDigit || alphaDigit;
}

/*********************************************************************
 * addDigit:
 *   Starting with currNum, adds on digit as the next digit in the
 *   number. So, for example, in base two, currNum might be 10 (i.e.,
 *   1010 base two) and digit might be '1'. So, we shift currNum over
 *   (10100 base two is 20) and add the value of digit (1) to get 21.
 *   Returns the resulting value.
 *
 *     currNum: the number so far
 *     digit: the next digit
 *
 *precondition: currNum >= 0 and digit is valid in BASE
 *postcondition: returns correct value for currNum with digit appended
 *********************************************************************/
int addDigit(int currNum, char digit)
{
  int digitValue;
  
  assert(currNum >= 0);
  assert(isValidDigit(digit));

  digitValue = getNumFromDigit(digit);
  
  /* If the number is too large to shift, then quit. */
  if (currNum > INT_MAX / BASE) {
    printf("That number is too large.\n");
    exit(EXIT_FAILURE);
  }

  /* Shift currNum over to make room for the new digit. */
  currNum *= BASE;

  /* If the number is too large to add the digit, then quit. */
  if (currNum > INT_MAX - digitValue) {
    printf("That number is too large.\n");
    exit(EXIT_FAILURE);
  }

  /* Add in the new digit. */
  currNum += digitValue;

  return currNum;
}

/*********************************************************************
 * getNumFromDigit:
 *   Given a digit in BASE, returns its numerical value.
 *
 *     digit: the digit to translate
 *
 *precondition: digit is valid in BASE
 *postcondition: returns correct value for digit in BASE
 *********************************************************************/
int getNumFromDigit(char digit)
{
  assert(isValidDigit(digit));

  /* Handle "normal" digits. */
  if (digit >= '0' && digit <= '9') {
    return (int)digit - (int)'0';
  }
  /* Handle alphabetical digits. */
  else {
    return (int)digit - (int)'A' + 10;
  }
}

/*********************************************************************
 * isPrime:
 *   Returns whether the given number is prime
 *
 *   Note: 1 is neither prime nor composite. So, isPrime returns 0
 *   for 1 (indicating not prime but NOT indicating composite).
 *
 *     num: the number to check for primality
 *
 *precondition: num > 0
 *postcondition: returns 1 if num is prime; otherwise, 0
 *********************************************************************/
int isPrime(int num)
{
  int limit = (int)sqrt(num);   /* The largest number to check. */
  int i;
  
  assert(num > 0);

  /* 1 is a special case: not prime. */
  if (num == 1) {
    return 0;
  }

  for (i = 2; i <= limit; i++) {
    /* Check for divisibility by i. */
    if (isDivisibleBy(num, i)) {
      return 0;
    }
  }

  /* num wasn't divisible by anything; so, it's prime. */
  return 1;
}

/*********************************************************************
 * isDivisibleBy:
 *   Returns whether num is divisible by divisor
 *
 *     num: the number to check for divisibility
 *     divisor: the number to check against
 *
 *precondition: num >= 0, divisor > 0
 *postcondition: returns 1 if num is divisible by divisor; else, 0
 *********************************************************************/
int isDivisibleBy(int num, int divisor)
{
  return num % divisor == 0;
}

